//
// gmtime_r.c
//
// Time routines
//
// Copyright (C) 2002 Michael Ringgaard. All rights reserved.
// Copyright (C) 2005 Jorgen Birkler
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright 
//    notice, this list of conditions and the following disclaimer.  
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.  
// 3. Neither the name of the project nor the names of its contributors
//    may be used to endorse or promote products derived from this software
//    without specific prior written permission. 
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
// SUCH DAMAGE.
// 


// AVR clib changes (mainly to reduce size, approx half original code):
//
// - changed to uint16_t from long where possible
// - replaced macros by static functions to keep size down
// - replaced summing of days from beginning of year in a array
// - Approximates year overflow with MAX_YEARS_APPROX instead of 
//   calculating exactly to the bit near of MAX_LONG
// - By buffer one year before overflow the seconds from month,day,hour,min,sec
//   can be safely added without checking for overflow.
//

#include <timeh_private.h>


struct tm *gmtime_r(const time_t *timer, struct tm *tmbuf)
{
  time_t time = *timer;
  uint16_t dayno, temp_days_size; //2147483647L => max 24856 days 
  uint16_t year = EPOCH_YR;
  uint32_t dayclock;  //86400 sec per day => need 32 bits
 
  dayno = time / SECS_DAY;
  dayclock = time % SECS_DAY;

  tmbuf->tm_sec = dayclock % 60;
  tmbuf->tm_min = (dayclock % 3600) / 60;
  tmbuf->tm_hour = dayclock / 3600;
  tmbuf->tm_wday = (dayno + 4) % 7; // Day 0 was a thursday
  while (dayno >= (temp_days_size = days_in_year(year))) 
  {
    dayno -= temp_days_size;
    year++;
  }
  tmbuf->tm_year = year - YEAR0;
  tmbuf->tm_yday = dayno;
  tmbuf->tm_mon = 0;
  while (dayno >= (temp_days_size = days_in_month(year,tmbuf->tm_mon))) 
  {
    dayno -= temp_days_size;
    tmbuf->tm_mon++;
  }
  tmbuf->tm_mday = dayno + 1;
  tmbuf->tm_isdst = 0;
  return tmbuf;
}

struct tm *localtime_r(const time_t *timer, struct tm *tmbuf)
{
  time_t t;
  t = *timer - _timezone;
  return gmtime_r(&t, tmbuf);
}

struct tm *gmtime(const time_t *timer)
{
  static struct tm static_tm;
  return gmtime_r(timer, &static_tm);
}

struct tm *localtime(const time_t *timer)
{
  static struct tm static_tm;
  return localtime_r(timer, &static_tm);
}

