//
// time.c
//
// Time routines
//
// Copyright (C) 2002 Michael Ringgaard. All rights reserved.
// Copyright (C) 2005 Jorgen Birkler
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright 
//    notice, this list of conditions and the following disclaimer.  
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.  
// 3. Neither the name of the project nor the names of its contributors
//    may be used to endorse or promote products derived from this software
//    without specific prior written permission. 
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
// SUCH DAMAGE.
// 


// AVR clib changes (mainly to reduce size, approx half original code):
//
// - changed to uint16_t from long where possible
// - replaced macros by static functions to keep size down
// - replaced summing of days from beginning of year in a array
// - Approximates year overflow with MAX_YEARS_APPROX instead of 
//   calculating exactly to the bit near of MAX_LONG
// - By buffer one year before overflow the seconds from month,day,hour,min,sec
//   can be safely added without checking for overflow.
//

#include <timeh_private.h>


static uint16_t days_from_year_start(uint16_t year, uint8_t month)
{
	uint16_t yday;
	//Sum of days since beginning of year for non-leap year
	static const prog_uint16_t _year_table_sum[12] = 
	{
	  31,                            31+28,                            31+28+31,
	 	31+28+31+30,                   31+28+31+30+31,                   31+28+31+30+31+30,
	 	31+28+31+30+31+30+31,          31+28+31+30+31+30+31+31,          31+28+31+30+31+30+31+31+30,
	 	31+28+31+30+31+30+31+31+30+31, 31+28+31+30+31+30+31+31+30+31+30, 31+28+31+30+31+30+31+31+30+31+30+31
	};
	yday = pgm_read_word_near(&_year_table_sum[month]);
 	if (month >= 1 && is_leap_year(year)) yday++;
 	return yday;
}


time_t mktime(struct tm *tmbuf)
{
	uint16_t tm_year;
  uint16_t month_day;
  //Normalize secs, min and hours
  tmbuf->tm_min += tmbuf->tm_sec / 60;
  tmbuf->tm_sec %= 60;
  if (tmbuf->tm_sec < 0) 
  {
    tmbuf->tm_sec += 60;
    tmbuf->tm_min--;
  }
  tmbuf->tm_hour += tmbuf->tm_min / 60;
  tmbuf->tm_min = tmbuf->tm_min % 60;
  if (tmbuf->tm_min < 0) 
  {
    tmbuf->tm_min += 60;
    tmbuf->tm_hour--;
  }
  month_day = tmbuf->tm_hour / 24;
  tmbuf->tm_hour= tmbuf->tm_hour % 24;
  if (tmbuf->tm_hour < 0) 
  {
    tmbuf->tm_hour += 24;
    month_day--;
  }
  tmbuf->tm_year += tmbuf->tm_mon / 12;
  tmbuf->tm_mon %= 12;
  if (tmbuf->tm_mon < 0) 
  {
    tmbuf->tm_mon += 12;
    tmbuf->tm_year--;
  }
  month_day += (tmbuf->tm_mday - 1);
  //Normalize negative day:
  while (month_day < 0) 
  {
    if(--tmbuf->tm_mon < 0) 
    {
      tmbuf->tm_year--;
      tmbuf->tm_mon = 11;
    }
    month_day += days_in_month(YEAR0 + tmbuf->tm_year,tmbuf->tm_mon);
  }
  //Normalize to big day
  {
  	uint8_t temp;
	  while (month_day >= (temp = days_in_month(YEAR0 + tmbuf->tm_year,tmbuf->tm_mon))) 
	  {
	    month_day -= temp; 
	    if (++(tmbuf->tm_mon) == 12) 
	    {
	      tmbuf->tm_mon = 0;
	      tmbuf->tm_year++;
	    }
	  }
  }
  tmbuf->tm_mday = month_day + 1;
  
  tm_year = tmbuf->tm_year + YEAR0;
  //Year within range?
  if (EPOCH_YR <= tm_year && tm_year < END_EPOCH_YR) 
  {
	  /*unsigned*/ int32_t seconds; //signed 32 bit
  	uint16_t day = 0; //since 1970
	  uint16_t yday;
  	
	  while (tm_year-- > EPOCH_YR)
	  {
	  	day += days_in_year(tm_year);
	  }
	  
  	yday = days_from_year_start(tm_year,tmbuf->tm_mon);
  		
	  yday += (tmbuf->tm_mday - 1);
	  day += yday;
	
	  tmbuf->tm_yday = yday;
	  tmbuf->tm_wday = (day + 4) % 7;   // Day 0 was thursday (4)
	
	  seconds = day;
	  seconds *= SECS_DAY;
	  
	  seconds += ((tmbuf->tm_hour * 60L) + tmbuf->tm_min) * 60L + tmbuf->tm_sec;
	
	  // Now adjust according to timezone and daylight saving time
	  seconds += _timezone;
	  if (tmbuf->tm_isdst) seconds -= _dstbias;

    //We check overflow simply (in case of strange time zone values maybe too simply)
	  if (seconds < 0) return (time_t) -1;
	
	  return (time_t) seconds;
  }
  else {
  	return -1;
  }
}



